<p>{% trans %}
    For details on what gets tracked, please refer to <a href="https://tracking.octoprint.org" target="_blank" rel="noopener noreferrer">tracking.octoprint.org</a>
    and also the <a href="https://tracking.octoprint.org/privacy" target="_blank" rel="noopener noreferrer">Privacy Policy at tracking.octoprint.org</a>.
{% endtrans %}</p>
