<div class="control-group" title="{{ _('Plugin blacklist URL. You should normally not have to change this.')|edq }}">
    <label class="control-label" for="settings-serverPluginBlacklistUrl">{{ _('Blacklist URL') }}</label>
    <div class="controls">
        <input type="text" class="input-block-level" data-bind="value: server_pluginBlacklist_url" id="settings-serverPluginBlacklistUrl">
    </div>
</div>
