<h3>{{ _("Server") }}</h3>
<form class="form-horizontal" onsubmit="return false;">
    {% include "snippets/settings/server/serverAllowFraming.jinja2" %}

    <h3>{{ _('Commands') }}</h3>

    {% include "snippets/settings/server/serverCommandServerRestart.jinja2" %}
    {% include "snippets/settings/server/serverCommandSystemRestart.jinja2" %}
    {% include "snippets/settings/server/serverCommandSystemShutdown.jinja2" %}

    <h3>{{ _('Connectivity check') }}</h3>

    {% include "snippets/settings/server/serverOnlineCheckDescription.jinja2" %}
    {% include "snippets/settings/server/serverOnlineCheck.jinja2" %}

    <h3>{{ _('Plugin blacklist processing') }}</h3>

    {% include "snippets/settings/server/serverPluginBlacklistDescription.jinja2" %}
    {% include "snippets/settings/server/serverPluginBlacklist.jinja2" %}

    <h3>{{ _('Debug options') }}</h3>

    {% include "snippets/settings/server/serverPluginTimings.jinja2" %}
</form>
