<div class="control-group" title="{{ _('Interval in which to check for internet connectivity')|edq }}">
    <label class="control-label" for="settings-serverOnlineCheckInterval">{{ _('Check interval') }}</label>
    <div class="controls">
        <span class="input-append">
            <input type="number" min="1" step="1" class="input-mini" data-bind="value: server_onlineCheck_interval" id="settings-serverOnlineCheckInterval">
            <span class="add-on">min</span>
        </span>
    </div>
</div>
