#workdialog .pre-output span{display:block}#workdialog .pre-output .message{font-weight:bold}#workdialog .pre-output .error{font-weight:bold;color:#990000}#workdialog .pre-output .stdout{color:#333333}#workdialog .pre-output .stderr{color:#990000}#workdialog .pre-output .call{color:#000099}