<div class="control-group" title="{{ _('Command to restart the system OctoPrint is running on')|edq }}">
    <label class="control-label" for="settings-systemRestartCommand">{{ _('Restart system') }}</label>
    <div class="controls">
        <input type="text" class="input-block-level" data-bind="value: server_commands_systemRestartCommand" id="settings-systemRestartCommand">
    </div>
</div>
