<div class="control-group">
    <div class="controls">
        <label class="checkbox">
            <input type="checkbox" data-bind="checked: settings.devel.pluginTimings" id="settings-develPluginTimings"> {{ _('Enable <code>plugintimings.log</code>') }} <label class="label label-important">{{ _("Warning") }}</label>
            <span class="help-block">{{ _('While this can negatively impact performance, a <code>plugintimings.log</code> can be incredibly useful for debugging any performance issues observed outside of safe mode.') }}</span>
        </label>
    </div>
</div>
