#workdialog {
  .pre-output {
    span {
      display: block; // so we don't need to add individual <br> tags to each line
    }

    .message {
      font-weight: bold;
    }

    .error {
      font-weight: bold;
      color: #990000;
    }

    .stdout {
      color: #333333;
    }

    .stderr {
      color: #990000;
    }

    .call {
      color: #000099;
    }
  }
}
